import os.path
from antlr4 import FileStream
from . import lines, print_to
from .antlr import Antlr

from .grammar.G4VisitorLexer import G4VisitorLexer
from .grammar.G4VisitorParser import G4VisitorParser
from .grammar.G4VVisitor import *


class G4VisitorVisitor(G4VVisitor):
    def visitAttributeRef(self, ctx:G4VisitorParser.AttributeRefContext):
        return super(G4VisitorVisitor, self).visitAttributeRef(ctx)[1:-1]


def generate_code(in_file, out_file=None):
    antlr = Antlr(G4VisitorLexer, G4VisitorParser)
    p = antlr.parser(FileStream(in_file))
    model = p.visitor().accept(G4VisitorVisitor())

    dirname = os.path.dirname(in_file)
    basename, ext = os.path.splitext(os.path.basename(in_file))
    if basename != model.name or ext != '.g4v':
        raise ValueError(f"Expected visitor in file '{model.name}.g4v', not in'{basename}{ext}'.")

    if out_file is None:
        out_file = os.path.join(dirname, basename + '.py')

    @print_to(out_file)
    def code():
        def cap(str):
            return str[0:1].upper() + str[1:]

        def visitor_code(visitor):
            visitorName, grammarName, rules = visitor

            yield from lines(f"""\
from gsl import pseudo_tuple

from gsl.antlr import ParseTreeVisitor
if __name__ is not None and "." in __name__:
    from .{grammarName}Parser import {grammarName}Parser
else:
    from {grammarName}Parser import {grammarName}Parser


""")
            for ruleName, body in rules:
                if isinstance(body, ObjectBody):
                    objectName, params = body
                    paramsStr = ' '.join(f"{param.name!r}," for param in params)
                    yield from lines(f"""\
{objectName} = pseudo_tuple({objectName!r}, ({paramsStr}))""")

            yield from lines(f"""\


class {visitorName}(ParseTreeVisitor):""")
            for ruleName, body in rules:
                yield from lines(f"""\
    def visit{cap(ruleName)}(self, ctx: {grammarName}Parser.{cap(ruleName)}Context):""")
                yield from body_code(body)
                yield from lines(f"""\

""")

        def body_code(body):
            if isinstance(body, ObjectBody):
                yield from object_body_code(body)
            else:
                yield from expr_body_code(body)

        def object_body_code(objectBody):
            yield from lines(f"""\
        return {objectBody.name}(""")
            for paramName, expr, optional in objectBody.params:
                opt = f" if {expr_core_str(expr, True)} else None" if optional else ""
                yield from lines(f"""\
            {expr_str(expr)}{opt},""")
            yield from lines(f"""\
        )""")

        def expr_body_code(exprBody):
            yield from lines(f"""\
        return {expr_str(exprBody)}""")

        def expr_core_str(expr, check=False):
            if isinstance(expr, RuleExpr):
                args = "ctx" + ''.join(f", {model.grammar}Parser.{cap(t)}Context" for t in expr.rules)
                operation = "self.has_children" if check else "self.get_children" if expr.multi else "self.get_child"
                return f"{operation}({args})"
            elif isinstance(expr, TokenExpr):
                return f"ctx.{expr.token}()"
            elif isinstance(expr, RefExpr):
                return f"ctx.{expr.ref}"

        def expr_str(expr):
            core = expr_core_str(expr)
            operation = "bool" if expr.presence else "self.visitNode" if isinstance(expr, RefExpr) or not expr.multi else "self.visitNodes"
            return f"{operation}({core})"

        yield from visitor_code(model)


def main():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('in_files', metavar='in_file', nargs='+')
    parser.add_argument('-o', '--out_file', default=None,
                        help="if there is only one input file, use this as the output file")

    args = parser.parse_args()

    if args.out_file is not None and len(args.in_files) > 1:
        print("explicit out_file only allowed for a single in_file")

    for in_file in args.in_files:
        generate_code(in_file, args.out_file)


if __name__ == '__main__':
    main()



from contextlib import contextmanager
import re
import threading


def pseudo_tuple(name, fields):
    if not re.match('^[_a-zA-Z][_a-zA-Z0-9]*$', name):
        raise ValueError(f"name must be a legal identifier: {name}")

    if len(fields) != len(set(fields)):
        raise ValueError("all fields must be distinct")

    for field in fields:
        if not re.match('^[a-zA-Z][_a-zA-Z0-9]*$', name):
            raise ValueError(f"name must be a legal identifier and not start with an underscore: {name}")

    def __init__(self, *args, **kwargs):
        if len(args) > len(fields):
            raise ValueError(f"extra positional args: {args[len(fields):]}")
        for field, arg in zip(fields, args):
            if field in kwargs:
                raise ValueError(f"argument duplicated in kwargs: {field}")
            kwargs[field] = arg
        for field in fields[len(args):]:
            if field not in kwargs:
                kwargs[field] = None
        self.__dict__.update(**kwargs)

    def __iter__(self):
        return (self.__dict__[field] for field in fields)

    def __str__(self):
        pos_items = [(k, self.__dict__[k]) for k in fields]
        kw_items = [(k, v) for k, v in self.__dict__.items() if k not in fields]
        args = ', '.join(f"{k}={v}" for k, v in pos_items + kw_items)
        return f"{name}({args})"

    def __repr__(self):
        pos_items = [(k, self.__dict__[k]) for k in fields]
        kw_items = [(k, v) for k, v in self.__dict__.items() if k not in fields]
        args = ', '.join(f"{k}={v!r}" for k, v in pos_items + kw_items)
        return f"{name}({args})"

    return type(name, (), {
        '__init__': __init__,
        '__iter__': __iter__,
        '__str__': __str__,
        '__repr__': __repr__,
    })


def lines(str):
    yield from str.splitlines()


def printlines(lines, end='\n', file=None):
    for line in lines:
        print(line, end=end, file=file)


def print_to(file, mode='w'):
    def decorator(fn):
        with open(file, mode) as f:
            printlines(fn(), file=f)
    return decorator


def generate(file):
    EMPTY, OPEN, CLOSE = 'empty', 'open', 'close'

    def marker_match(line):
        return re.search(r'<(default )?GSL customizable: ([-\w]+)( /)?>|</GSL customizable: ([-\w]+)>', line)

    def marker_info(m):
        default = m.group(1) is not None
        opening = m.group(2) is not None
        self_closing = m.group(3) is not None
        name = m.group(2) if opening else m.group(4)

        return name, EMPTY if self_closing else OPEN if opening else CLOSE, default if opening else None

    def marker_format(m, name, mode, default):
        prefix = '/' if mode == CLOSE else 'default ' if default else ''
        suffix = ' /' if mode == EMPTY else ''
        return f'{m.string[:m.start()]}<{prefix}GSL customizable: {name}{suffix}>{m.string[m.end():]}'

    def decorator(fn):
        sections = {}
        try:
            f = open(file)
        except FileNotFoundError:
            pass
        else:
            with f:
                open_section = None
                for i, line in enumerate(f, start=1):
                    m = marker_match(line)
                    if m:
                        name, mode, default = marker_info(m)
                        if mode == CLOSE:
                            if not open_section:
                                raise ValueError(f"Line {i} (old): closing unopened customizable '{name}'")
                            elif name != open_section:
                                raise ValueError(f"Line {i} (old): closing unopened customizable '{name}'"
                                                 f" (open customizable is '{open_section}')")

                            open_section = None
                        elif mode == OPEN:
                            if open_section:
                                raise ValueError(f"Line {i} (old): nested customizable '{name}'")
                            if name in sections:
                                raise ValueError(f"Line {i} (old): duplicate customizable '{name}'")
                            if not default:
                                sections[name] = []

                            open_section = name
                        elif mode == EMPTY:
                            if open_section:
                                raise ValueError(f"Line {i} (old): nested customizable '{name}'")
                            if name in sections:
                                raise ValueError(f"Line {i} (old): duplicate customizable '{name}'")
                            if not default:
                                sections[name] = None
                    elif open_section and open_section in sections:
                        sections[open_section].append(line.rstrip('\r\n'))
                if open_section:
                    raise ValueError(f"Line {i} (old): unclosed customizable '{open_section}'")

        @print_to(file)
        def code():
            open_section = None
            new_sections = set()
            skip = False
            for i, line in enumerate(fn(), start=1):
                m = marker_match(line)
                if m:
                    name, mode, default = marker_info(m)
                    if mode == CLOSE:
                        if not open_section:
                            raise ValueError(f"Line {i} (new): closing unopened customizable '{name}'")
                        elif name != open_section:
                            raise ValueError(f"Line {i} (new): closing unopened customizable '{name}'"
                                             f" (open customizable is '{open_section}')")

                        if not skip:
                            yield marker_format(m, name, CLOSE, None)

                        open_section = None
                        skip = False
                    elif mode == OPEN:
                        if open_section:
                            raise ValueError(f"Line {i} (new): nested customizable '{name}'")
                        if name in new_sections:
                            raise ValueError(f"Line {i} (new): duplicate customizable '{name}'")
                        if not default:
                            raise ValueError(f"Line {i} (new): generated code must declare sections as `default`")
                        new_sections.add(name)

                        if name in sections:
                            if sections[name] is None:
                                yield marker_format(m, name, EMPTY, False)
                            else:
                                yield marker_format(m, name, OPEN, False)
                                yield from sections[name]
                                yield marker_format(m, name, CLOSE, None)
                        else:
                            yield marker_format(m, name, OPEN, True)

                        open_section = name
                        skip = name in sections
                    elif mode == EMPTY:
                        if open_section:
                            raise ValueError(f"Line {i} (new): nested customizable '{name}'")
                        if name in new_sections:
                            raise ValueError(f"Line {i} (new): duplicate customizable '{name}'")
                        if not default:
                            raise ValueError(f"Line {i} (new): generated code must declare sections as `default`")
                        new_sections.add(name)

                        if name in sections:
                            if sections[name] is None:
                                yield marker_format(m, name, EMPTY, False)
                            else:
                                yield marker_format(m, name, OPEN, False)
                                yield from sections[name]
                                yield marker_format(m, name, CLOSE, None)
                        else:
                            yield marker_format(m, name, EMPTY, True)
                elif not skip:
                    yield line
            if open_section:
                raise ValueError(f"Line {i} (new): unclosed customizable '{open_section}'")

    return decorator



from collections import namedtuple

from gsl.antlr import ParseTreeVisitor
from .SetTestParser import SetTestParser




class SetTestVisitor(ParseTreeVisitor):
    def visitExpr(self, ctx: SetTestParser.ExprContext):
        return self.visitNode(self.get_child(ctx, SetTestParser.Set0Context))

    def visitSet0(self, ctx: SetTestParser.Set0Context):
        return self.visitNodes(self.get_children(ctx, SetTestParser.ElementContext))

    def visitElement(self, ctx: SetTestParser.ElementContext):
        return self.visitNode(self.get_child(ctx))

    def visitIntElement(self, ctx: SetTestParser.IntElementContext):
        return self.visitNode(self.get_child(ctx))




class DotDict(dict):
    def __getattribute__(self, k):
        try:
            v = self[k]
        except KeyError:
            return super().__getattribute__(k)
        return v

    def __setattr__(self, k, v):
        try:
            self[k] = v
        except KeyError:
            return super().__setattr__(k, v)



# Generated from tests/grammar/HedgehogTest.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\34")
        buf.write("\u00b1\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\4\17\t\17\3\2\7\2 \n\2\f\2\16\2#\13\2\3\2\3\2\3")
        buf.write("\3\3\3\3\3\3\3\3\3\3\3\5\3-\n\3\3\3\3\3\7\3\61\n\3\f\3")
        buf.write("\16\3\64\13\3\3\3\7\3\67\n\3\f\3\16\3:\13\3\3\3\3\3\3")
        buf.write("\4\5\4?\n\4\3\4\3\4\3\4\3\4\3\4\3\4\7\4G\n\4\f\4\16\4")
        buf.write("J\13\4\3\4\3\4\3\5\3\5\3\5\3\5\7\5R\n\5\f\5\16\5U\13\5")
        buf.write("\3\5\3\5\3\6\3\6\3\6\3\6\3\6\5\6^\n\6\3\6\3\6\3\6\3\6")
        buf.write("\3\6\3\6\5\6f\n\6\3\6\5\6i\n\6\3\7\3\7\3\7\3\7\3\7\3\7")
        buf.write("\5\7q\n\7\3\7\3\7\3\b\3\b\3\b\5\bx\n\b\3\b\3\b\5\b|\n")
        buf.write("\b\3\b\3\b\3\b\5\b\u0081\n\b\3\b\3\b\5\b\u0085\n\b\3\t")
        buf.write("\3\t\3\t\7\t\u008a\n\t\f\t\16\t\u008d\13\t\3\n\3\n\3\n")
        buf.write("\3\13\3\13\3\13\3\13\7\13\u0096\n\13\f\13\16\13\u0099")
        buf.write("\13\13\3\13\3\13\3\f\3\f\3\f\7\f\u00a0\n\f\f\f\16\f\u00a3")
        buf.write("\13\f\3\r\3\r\3\r\7\r\u00a8\n\r\f\r\16\r\u00ab\13\r\3")
        buf.write("\16\3\16\3\17\3\17\3\17\2\2\20\2\4\6\b\n\f\16\20\22\24")
        buf.write("\26\30\32\34\2\3\3\2\20\22\2\u00b8\2!\3\2\2\2\4&\3\2\2")
        buf.write("\2\6>\3\2\2\2\bM\3\2\2\2\nh\3\2\2\2\fj\3\2\2\2\16\u0084")
        buf.write("\3\2\2\2\20\u0086\3\2\2\2\22\u008e\3\2\2\2\24\u0091\3")
        buf.write("\2\2\2\26\u009c\3\2\2\2\30\u00a4\3\2\2\2\32\u00ac\3\2")
        buf.write("\2\2\34\u00ae\3\2\2\2\36 \5\4\3\2\37\36\3\2\2\2 #\3\2")
        buf.write("\2\2!\37\3\2\2\2!\"\3\2\2\2\"$\3\2\2\2#!\3\2\2\2$%\7\2")
        buf.write("\2\3%\3\3\2\2\2&\'\5\30\r\2\'(\5\32\16\2()\7\16\2\2)*")
        buf.write("\5\34\17\2*,\7\27\2\2+-\7\7\2\2,+\3\2\2\2,-\3\2\2\2-\62")
        buf.write("\3\2\2\2.\61\5\6\4\2/\61\5\b\5\2\60.\3\2\2\2\60/\3\2\2")
        buf.write("\2\61\64\3\2\2\2\62\60\3\2\2\2\62\63\3\2\2\2\638\3\2\2")
        buf.write("\2\64\62\3\2\2\2\65\67\5\f\7\2\66\65\3\2\2\2\67:\3\2\2")
        buf.write("\28\66\3\2\2\289\3\2\2\29;\3\2\2\2:8\3\2\2\2;<\7\30\2")
        buf.write("\2<\5\3\2\2\2=?\7\3\2\2>=\3\2\2\2>?\3\2\2\2?@\3\2\2\2")
        buf.write("@A\5\32\16\2AB\5\32\16\2BC\7\16\2\2CD\5\34\17\2DH\7\27")
        buf.write("\2\2EG\5\n\6\2FE\3\2\2\2GJ\3\2\2\2HF\3\2\2\2HI\3\2\2\2")
        buf.write("IK\3\2\2\2JH\3\2\2\2KL\7\30\2\2L\7\3\2\2\2MN\7\4\2\2N")
        buf.write("O\5\32\16\2OS\7\27\2\2PR\5\6\4\2QP\3\2\2\2RU\3\2\2\2S")
        buf.write("Q\3\2\2\2ST\3\2\2\2TV\3\2\2\2US\3\2\2\2VW\7\30\2\2W\t")
        buf.write("\3\2\2\2XY\7\5\2\2YZ\7\13\2\2Z]\7\b\2\2[\\\7\n\2\2\\^")
        buf.write("\7\b\2\2][\3\2\2\2]^\3\2\2\2^_\3\2\2\2_i\7\f\2\2`a\7\6")
        buf.write("\2\2ab\7\13\2\2be\7\b\2\2cd\7\n\2\2df\7\b\2\2ec\3\2\2")
        buf.write("\2ef\3\2\2\2fg\3\2\2\2gi\7\f\2\2hX\3\2\2\2h`\3\2\2\2i")
        buf.write("\13\3\2\2\2jk\t\2\2\2kl\5\30\r\2lm\7\23\2\2mn\5\16\b\2")
        buf.write("np\7\24\2\2oq\7\7\2\2po\3\2\2\2pq\3\2\2\2qr\3\2\2\2rs")
        buf.write("\7\f\2\2s\r\3\2\2\2tw\5\20\t\2uv\7\n\2\2vx\5\22\n\2wu")
        buf.write("\3\2\2\2wx\3\2\2\2x{\3\2\2\2yz\7\n\2\2z|\5\24\13\2{y\3")
        buf.write("\2\2\2{|\3\2\2\2|\u0085\3\2\2\2}\u0080\5\22\n\2~\177\7")
        buf.write("\n\2\2\177\u0081\5\24\13\2\u0080~\3\2\2\2\u0080\u0081")
        buf.write("\3\2\2\2\u0081\u0085\3\2\2\2\u0082\u0085\5\24\13\2\u0083")
        buf.write("\u0085\3\2\2\2\u0084t\3\2\2\2\u0084}\3\2\2\2\u0084\u0082")
        buf.write("\3\2\2\2\u0084\u0083\3\2\2\2\u0085\17\3\2\2\2\u0086\u008b")
        buf.write("\5\32\16\2\u0087\u0088\7\n\2\2\u0088\u008a\5\32\16\2\u0089")
        buf.write("\u0087\3\2\2\2\u008a\u008d\3\2\2\2\u008b\u0089\3\2\2\2")
        buf.write("\u008b\u008c\3\2\2\2\u008c\21\3\2\2\2\u008d\u008b\3\2")
        buf.write("\2\2\u008e\u008f\7\17\2\2\u008f\u0090\5\32\16\2\u0090")
        buf.write("\23\3\2\2\2\u0091\u0092\7\25\2\2\u0092\u0097\5\26\f\2")
        buf.write("\u0093\u0094\7\n\2\2\u0094\u0096\5\26\f\2\u0095\u0093")
        buf.write("\3\2\2\2\u0096\u0099\3\2\2\2\u0097\u0095\3\2\2\2\u0097")
        buf.write("\u0098\3\2\2\2\u0098\u009a\3\2\2\2\u0099\u0097\3\2\2\2")
        buf.write("\u009a\u009b\7\26\2\2\u009b\25\3\2\2\2\u009c\u00a1\5\32")
        buf.write("\16\2\u009d\u009e\7\r\2\2\u009e\u00a0\5\32\16\2\u009f")
        buf.write("\u009d\3\2\2\2\u00a0\u00a3\3\2\2\2\u00a1\u009f\3\2\2\2")
        buf.write("\u00a1\u00a2\3\2\2\2\u00a2\27\3\2\2\2\u00a3\u00a1\3\2")
        buf.write("\2\2\u00a4\u00a9\5\32\16\2\u00a5\u00a6\7\t\2\2\u00a6\u00a8")
        buf.write("\5\32\16\2\u00a7\u00a5\3\2\2\2\u00a8\u00ab\3\2\2\2\u00a9")
        buf.write("\u00a7\3\2\2\2\u00a9\u00aa\3\2\2\2\u00aa\31\3\2\2\2\u00ab")
        buf.write("\u00a9\3\2\2\2\u00ac\u00ad\7\31\2\2\u00ad\33\3\2\2\2\u00ae")
        buf.write("\u00af\7\32\2\2\u00af\35\3\2\2\2\26!,\60\628>HS]ehpw{")
        buf.write("\u0080\u0084\u008b\u0097\u00a1\u00a9")
        return buf.getvalue()


class HedgehogTestParser ( Parser ):

    grammarFileName = "HedgehogTest.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'repeated'", "'oneof'", "'Python'", "'TypeScript'", 
                     "<INVALID>", "<INVALID>", "'.'", "','", "':'", "';'", 
                     "'/'", "'='", "'*'", "'=>'", "'<='", "'<-'", "'('", 
                     "')'", "'['", "']'", "'{'", "'}'" ]

    symbolicNames = [ "<INVALID>", "<INVALID>", "<INVALID>", "<INVALID>", 
                      "<INVALID>", "DOCSTRING", "STRING", "DOT", "COMMA", 
                      "COLON", "SEMI", "SLASH", "EQUALS", "STAR", "REQ", 
                      "REP", "UPD", "LPAREN", "RPAREN", "LBRACKET", "RBRACKET", 
                      "LBRACE", "RBRACE", "IDENTIFIER", "NUMBER", "COMMENT", 
                      "WS" ]

    RULE_expr = 0
    RULE_message = 1
    RULE_field = 2
    RULE_oneof = 3
    RULE_languageFieldSpec = 4
    RULE_messageClass = 5
    RULE_paramList = 6
    RULE_mandatoryParamList = 7
    RULE_repeatedParam = 8
    RULE_optionalParamList = 9
    RULE_optionalParam = 10
    RULE_qualifiedIdentifier = 11
    RULE_identifier = 12
    RULE_number = 13

    ruleNames =  [ "expr", "message", "field", "oneof", "languageFieldSpec", 
                   "messageClass", "paramList", "mandatoryParamList", "repeatedParam", 
                   "optionalParamList", "optionalParam", "qualifiedIdentifier", 
                   "identifier", "number" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    DOCSTRING=5
    STRING=6
    DOT=7
    COMMA=8
    COLON=9
    SEMI=10
    SLASH=11
    EQUALS=12
    STAR=13
    REQ=14
    REP=15
    UPD=16
    LPAREN=17
    RPAREN=18
    LBRACKET=19
    RBRACKET=20
    LBRACE=21
    RBRACE=22
    IDENTIFIER=23
    NUMBER=24
    COMMENT=25
    WS=26

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(HedgehogTestParser.EOF, 0)

        def message(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.MessageContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.MessageContext,i)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_expr

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpr" ):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)




    def expr(self):

        localctx = HedgehogTestParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_expr)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 31
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.IDENTIFIER:
                self.state = 28
                self.message()
                self.state = 33
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 34
            self.match(HedgehogTestParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class MessageContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.messageType = None # QualifiedIdentifierContext
            self.discriminator = None # IdentifierContext
            self.label = None # NumberContext

        def EQUALS(self):
            return self.getToken(HedgehogTestParser.EQUALS, 0)

        def LBRACE(self):
            return self.getToken(HedgehogTestParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(HedgehogTestParser.RBRACE, 0)

        def qualifiedIdentifier(self):
            return self.getTypedRuleContext(HedgehogTestParser.QualifiedIdentifierContext,0)


        def identifier(self):
            return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,0)


        def number(self):
            return self.getTypedRuleContext(HedgehogTestParser.NumberContext,0)


        def DOCSTRING(self):
            return self.getToken(HedgehogTestParser.DOCSTRING, 0)

        def field(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.FieldContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.FieldContext,i)


        def oneof(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.OneofContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.OneofContext,i)


        def messageClass(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.MessageClassContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.MessageClassContext,i)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_message

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitMessage" ):
                return visitor.visitMessage(self)
            else:
                return visitor.visitChildren(self)




    def message(self):

        localctx = HedgehogTestParser.MessageContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_message)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 36
            localctx.messageType = self.qualifiedIdentifier()
            self.state = 37
            localctx.discriminator = self.identifier()
            self.state = 38
            self.match(HedgehogTestParser.EQUALS)
            self.state = 39
            localctx.label = self.number()
            self.state = 40
            self.match(HedgehogTestParser.LBRACE)
            self.state = 42
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==HedgehogTestParser.DOCSTRING:
                self.state = 41
                self.match(HedgehogTestParser.DOCSTRING)


            self.state = 48
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << HedgehogTestParser.T__0) | (1 << HedgehogTestParser.T__1) | (1 << HedgehogTestParser.IDENTIFIER))) != 0):
                self.state = 46
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [HedgehogTestParser.T__0, HedgehogTestParser.IDENTIFIER]:
                    self.state = 44
                    self.field()
                    pass
                elif token in [HedgehogTestParser.T__1]:
                    self.state = 45
                    self.oneof()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 50
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 54
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << HedgehogTestParser.REQ) | (1 << HedgehogTestParser.REP) | (1 << HedgehogTestParser.UPD))) != 0):
                self.state = 51
                self.messageClass()
                self.state = 56
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 57
            self.match(HedgehogTestParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FieldContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.rep = None # Token
            self.fieldType = None # IdentifierContext
            self.name = None # IdentifierContext
            self.label = None # NumberContext

        def EQUALS(self):
            return self.getToken(HedgehogTestParser.EQUALS, 0)

        def LBRACE(self):
            return self.getToken(HedgehogTestParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(HedgehogTestParser.RBRACE, 0)

        def identifier(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.IdentifierContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,i)


        def number(self):
            return self.getTypedRuleContext(HedgehogTestParser.NumberContext,0)


        def languageFieldSpec(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.LanguageFieldSpecContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.LanguageFieldSpecContext,i)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_field

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitField" ):
                return visitor.visitField(self)
            else:
                return visitor.visitChildren(self)




    def field(self):

        localctx = HedgehogTestParser.FieldContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_field)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 60
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==HedgehogTestParser.T__0:
                self.state = 59
                localctx.rep = self.match(HedgehogTestParser.T__0)


            self.state = 62
            localctx.fieldType = self.identifier()
            self.state = 63
            localctx.name = self.identifier()
            self.state = 64
            self.match(HedgehogTestParser.EQUALS)
            self.state = 65
            localctx.label = self.number()
            self.state = 66
            self.match(HedgehogTestParser.LBRACE)
            self.state = 70
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.T__2 or _la==HedgehogTestParser.T__3:
                self.state = 67
                self.languageFieldSpec()
                self.state = 72
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 73
            self.match(HedgehogTestParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OneofContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.name = None # IdentifierContext

        def LBRACE(self):
            return self.getToken(HedgehogTestParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(HedgehogTestParser.RBRACE, 0)

        def identifier(self):
            return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,0)


        def field(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.FieldContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.FieldContext,i)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_oneof

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOneof" ):
                return visitor.visitOneof(self)
            else:
                return visitor.visitChildren(self)




    def oneof(self):

        localctx = HedgehogTestParser.OneofContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_oneof)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 75
            self.match(HedgehogTestParser.T__1)
            self.state = 76
            localctx.name = self.identifier()
            self.state = 77
            self.match(HedgehogTestParser.LBRACE)
            self.state = 81
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.T__0 or _la==HedgehogTestParser.IDENTIFIER:
                self.state = 78
                self.field()
                self.state = 83
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 84
            self.match(HedgehogTestParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LanguageFieldSpecContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_languageFieldSpec

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class TypeScriptSpecContext(LanguageFieldSpecContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a HedgehogTestParser.LanguageFieldSpecContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def COLON(self):
            return self.getToken(HedgehogTestParser.COLON, 0)
        def STRING(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.STRING)
            else:
                return self.getToken(HedgehogTestParser.STRING, i)
        def SEMI(self):
            return self.getToken(HedgehogTestParser.SEMI, 0)
        def COMMA(self):
            return self.getToken(HedgehogTestParser.COMMA, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTypeScriptSpec" ):
                return visitor.visitTypeScriptSpec(self)
            else:
                return visitor.visitChildren(self)


    class PythonSpecContext(LanguageFieldSpecContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a HedgehogTestParser.LanguageFieldSpecContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def COLON(self):
            return self.getToken(HedgehogTestParser.COLON, 0)
        def STRING(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.STRING)
            else:
                return self.getToken(HedgehogTestParser.STRING, i)
        def SEMI(self):
            return self.getToken(HedgehogTestParser.SEMI, 0)
        def COMMA(self):
            return self.getToken(HedgehogTestParser.COMMA, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitPythonSpec" ):
                return visitor.visitPythonSpec(self)
            else:
                return visitor.visitChildren(self)



    def languageFieldSpec(self):

        localctx = HedgehogTestParser.LanguageFieldSpecContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_languageFieldSpec)
        self._la = 0 # Token type
        try:
            self.state = 102
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [HedgehogTestParser.T__2]:
                localctx = HedgehogTestParser.PythonSpecContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 86
                self.match(HedgehogTestParser.T__2)
                self.state = 87
                self.match(HedgehogTestParser.COLON)
                self.state = 88
                self.match(HedgehogTestParser.STRING)
                self.state = 91
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==HedgehogTestParser.COMMA:
                    self.state = 89
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 90
                    self.match(HedgehogTestParser.STRING)


                self.state = 93
                self.match(HedgehogTestParser.SEMI)
                pass
            elif token in [HedgehogTestParser.T__3]:
                localctx = HedgehogTestParser.TypeScriptSpecContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 94
                self.match(HedgehogTestParser.T__3)
                self.state = 95
                self.match(HedgehogTestParser.COLON)
                self.state = 96
                self.match(HedgehogTestParser.STRING)
                self.state = 99
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==HedgehogTestParser.COMMA:
                    self.state = 97
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 98
                    self.match(HedgehogTestParser.STRING)


                self.state = 101
                self.match(HedgehogTestParser.SEMI)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class MessageClassContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def qualifiedIdentifier(self):
            return self.getTypedRuleContext(HedgehogTestParser.QualifiedIdentifierContext,0)


        def LPAREN(self):
            return self.getToken(HedgehogTestParser.LPAREN, 0)

        def paramList(self):
            return self.getTypedRuleContext(HedgehogTestParser.ParamListContext,0)


        def RPAREN(self):
            return self.getToken(HedgehogTestParser.RPAREN, 0)

        def SEMI(self):
            return self.getToken(HedgehogTestParser.SEMI, 0)

        def REQ(self):
            return self.getToken(HedgehogTestParser.REQ, 0)

        def REP(self):
            return self.getToken(HedgehogTestParser.REP, 0)

        def UPD(self):
            return self.getToken(HedgehogTestParser.UPD, 0)

        def DOCSTRING(self):
            return self.getToken(HedgehogTestParser.DOCSTRING, 0)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_messageClass

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitMessageClass" ):
                return visitor.visitMessageClass(self)
            else:
                return visitor.visitChildren(self)




    def messageClass(self):

        localctx = HedgehogTestParser.MessageClassContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_messageClass)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 104
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << HedgehogTestParser.REQ) | (1 << HedgehogTestParser.REP) | (1 << HedgehogTestParser.UPD))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
            self.state = 105
            self.qualifiedIdentifier()
            self.state = 106
            self.match(HedgehogTestParser.LPAREN)
            self.state = 107
            self.paramList()
            self.state = 108
            self.match(HedgehogTestParser.RPAREN)
            self.state = 110
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==HedgehogTestParser.DOCSTRING:
                self.state = 109
                self.match(HedgehogTestParser.DOCSTRING)


            self.state = 112
            self.match(HedgehogTestParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ParamListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def mandatoryParamList(self):
            return self.getTypedRuleContext(HedgehogTestParser.MandatoryParamListContext,0)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.COMMA)
            else:
                return self.getToken(HedgehogTestParser.COMMA, i)

        def repeatedParam(self):
            return self.getTypedRuleContext(HedgehogTestParser.RepeatedParamContext,0)


        def optionalParamList(self):
            return self.getTypedRuleContext(HedgehogTestParser.OptionalParamListContext,0)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_paramList

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitParamList" ):
                return visitor.visitParamList(self)
            else:
                return visitor.visitChildren(self)




    def paramList(self):

        localctx = HedgehogTestParser.ParamListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_paramList)
        self._la = 0 # Token type
        try:
            self.state = 130
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [HedgehogTestParser.IDENTIFIER]:
                self.enterOuterAlt(localctx, 1)
                self.state = 114
                self.mandatoryParamList()
                self.state = 117
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,12,self._ctx)
                if la_ == 1:
                    self.state = 115
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 116
                    self.repeatedParam()


                self.state = 121
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==HedgehogTestParser.COMMA:
                    self.state = 119
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 120
                    self.optionalParamList()


                pass
            elif token in [HedgehogTestParser.STAR]:
                self.enterOuterAlt(localctx, 2)
                self.state = 123
                self.repeatedParam()
                self.state = 126
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==HedgehogTestParser.COMMA:
                    self.state = 124
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 125
                    self.optionalParamList()


                pass
            elif token in [HedgehogTestParser.LBRACKET]:
                self.enterOuterAlt(localctx, 3)
                self.state = 128
                self.optionalParamList()
                pass
            elif token in [HedgehogTestParser.RPAREN]:
                self.enterOuterAlt(localctx, 4)

                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class MandatoryParamListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.IdentifierContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.COMMA)
            else:
                return self.getToken(HedgehogTestParser.COMMA, i)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_mandatoryParamList

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitMandatoryParamList" ):
                return visitor.visitMandatoryParamList(self)
            else:
                return visitor.visitChildren(self)




    def mandatoryParamList(self):

        localctx = HedgehogTestParser.MandatoryParamListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_mandatoryParamList)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 132
            self.identifier()
            self.state = 137
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,16,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 133
                    self.match(HedgehogTestParser.COMMA)
                    self.state = 134
                    self.identifier() 
                self.state = 139
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RepeatedParamContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def STAR(self):
            return self.getToken(HedgehogTestParser.STAR, 0)

        def identifier(self):
            return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,0)


        def getRuleIndex(self):
            return HedgehogTestParser.RULE_repeatedParam

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRepeatedParam" ):
                return visitor.visitRepeatedParam(self)
            else:
                return visitor.visitChildren(self)




    def repeatedParam(self):

        localctx = HedgehogTestParser.RepeatedParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_repeatedParam)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 140
            self.match(HedgehogTestParser.STAR)
            self.state = 141
            self.identifier()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptionalParamListContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACKET(self):
            return self.getToken(HedgehogTestParser.LBRACKET, 0)

        def optionalParam(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.OptionalParamContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.OptionalParamContext,i)


        def RBRACKET(self):
            return self.getToken(HedgehogTestParser.RBRACKET, 0)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.COMMA)
            else:
                return self.getToken(HedgehogTestParser.COMMA, i)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_optionalParamList

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOptionalParamList" ):
                return visitor.visitOptionalParamList(self)
            else:
                return visitor.visitChildren(self)




    def optionalParamList(self):

        localctx = HedgehogTestParser.OptionalParamListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_optionalParamList)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 143
            self.match(HedgehogTestParser.LBRACKET)
            self.state = 144
            self.optionalParam()
            self.state = 149
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.COMMA:
                self.state = 145
                self.match(HedgehogTestParser.COMMA)
                self.state = 146
                self.optionalParam()
                self.state = 151
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 152
            self.match(HedgehogTestParser.RBRACKET)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class OptionalParamContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.IdentifierContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,i)


        def SLASH(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.SLASH)
            else:
                return self.getToken(HedgehogTestParser.SLASH, i)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_optionalParam

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitOptionalParam" ):
                return visitor.visitOptionalParam(self)
            else:
                return visitor.visitChildren(self)




    def optionalParam(self):

        localctx = HedgehogTestParser.OptionalParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_optionalParam)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 154
            self.identifier()
            self.state = 159
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.SLASH:
                self.state = 155
                self.match(HedgehogTestParser.SLASH)
                self.state = 156
                self.identifier()
                self.state = 161
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class QualifiedIdentifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(HedgehogTestParser.IdentifierContext)
            else:
                return self.getTypedRuleContext(HedgehogTestParser.IdentifierContext,i)


        def DOT(self, i:int=None):
            if i is None:
                return self.getTokens(HedgehogTestParser.DOT)
            else:
                return self.getToken(HedgehogTestParser.DOT, i)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_qualifiedIdentifier

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitQualifiedIdentifier" ):
                return visitor.visitQualifiedIdentifier(self)
            else:
                return visitor.visitChildren(self)




    def qualifiedIdentifier(self):

        localctx = HedgehogTestParser.QualifiedIdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_qualifiedIdentifier)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 162
            self.identifier()
            self.state = 167
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==HedgehogTestParser.DOT:
                self.state = 163
                self.match(HedgehogTestParser.DOT)
                self.state = 164
                self.identifier()
                self.state = 169
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class IdentifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def IDENTIFIER(self):
            return self.getToken(HedgehogTestParser.IDENTIFIER, 0)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_identifier

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIdentifier" ):
                return visitor.visitIdentifier(self)
            else:
                return visitor.visitChildren(self)




    def identifier(self):

        localctx = HedgehogTestParser.IdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_identifier)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 170
            self.match(HedgehogTestParser.IDENTIFIER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NumberContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def NUMBER(self):
            return self.getToken(HedgehogTestParser.NUMBER, 0)

        def getRuleIndex(self):
            return HedgehogTestParser.RULE_number

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumber" ):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)




    def number(self):

        localctx = HedgehogTestParser.NumberContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_number)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 172
            self.match(HedgehogTestParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








# Generated from gsl/grammar/G4Visitor.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\31")
        buf.write("\u00c8\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t\7")
        buf.write("\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r\4\16")
        buf.write("\t\16\3\2\3\2\3\2\3\2\3\2\3\2\3\2\7\2$\n\2\f\2\16\2\'")
        buf.write("\13\2\3\2\3\2\3\3\3\3\3\3\3\3\3\3\3\4\3\4\3\4\3\4\3\4")
        buf.write("\3\4\7\4\66\n\4\f\4\16\49\13\4\3\4\5\4<\n\4\5\4>\n\4\3")
        buf.write("\4\3\4\3\4\3\4\3\4\3\4\7\4F\n\4\f\4\16\4I\13\4\3\4\5\4")
        buf.write("L\n\4\5\4N\n\4\3\4\3\4\3\4\3\4\3\4\7\4U\n\4\f\4\16\4X")
        buf.write("\13\4\3\4\5\4[\n\4\5\4]\n\4\3\4\5\4`\n\4\3\5\3\5\3\5\3")
        buf.write("\5\3\5\3\5\3\5\3\5\3\5\3\5\5\5l\n\5\3\6\3\6\3\6\3\6\3")
        buf.write("\6\5\6s\n\6\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\3\7\5")
        buf.write("\7\177\n\7\3\b\3\b\3\b\3\b\3\b\3\b\6\b\u0087\n\b\r\b\16")
        buf.write("\b\u0088\3\b\3\b\3\b\3\b\3\b\6\b\u0090\n\b\r\b\16\b\u0091")
        buf.write("\5\b\u0094\n\b\3\b\3\b\3\b\3\b\3\b\3\b\6\b\u009c\n\b\r")
        buf.write("\b\16\b\u009d\3\b\3\b\5\b\u00a2\n\b\3\b\3\b\5\b\u00a6")
        buf.write("\n\b\3\b\5\b\u00a9\n\b\5\b\u00ab\n\b\3\b\3\b\5\b\u00af")
        buf.write("\n\b\3\b\5\b\u00b2\n\b\3\b\3\b\5\b\u00b6\n\b\5\b\u00b8")
        buf.write("\n\b\3\t\3\t\3\n\3\n\3\13\3\13\5\13\u00c0\n\13\3\f\3\f")
        buf.write("\3\r\3\r\3\16\3\16\3\16\2\2\17\2\4\6\b\n\f\16\20\22\24")
        buf.write("\26\30\32\2\3\4\2\3\5\24\24\2\u00db\2\34\3\2\2\2\4*\3")
        buf.write("\2\2\2\6_\3\2\2\2\bk\3\2\2\2\nr\3\2\2\2\f~\3\2\2\2\16")
        buf.write("\u00b7\3\2\2\2\20\u00b9\3\2\2\2\22\u00bb\3\2\2\2\24\u00bf")
        buf.write("\3\2\2\2\26\u00c1\3\2\2\2\30\u00c3\3\2\2\2\32\u00c5\3")
        buf.write("\2\2\2\34\35\7\3\2\2\35\36\5\24\13\2\36\37\7\4\2\2\37")
        buf.write(" \7\5\2\2 !\5\24\13\2!%\7\13\2\2\"$\5\4\3\2#\"\3\2\2\2")
        buf.write("$\'\3\2\2\2%#\3\2\2\2%&\3\2\2\2&(\3\2\2\2\'%\3\2\2\2(")
        buf.write(")\7\2\2\3)\3\3\2\2\2*+\5\26\f\2+,\7\f\2\2,-\5\6\4\2-.")
        buf.write("\7\13\2\2.\5\3\2\2\2/`\5\16\b\2\60\61\5\24\13\2\61=\7")
        buf.write("\16\2\2\62\67\5\b\5\2\63\64\7\6\2\2\64\66\5\b\5\2\65\63")
        buf.write("\3\2\2\2\669\3\2\2\2\67\65\3\2\2\2\678\3\2\2\28;\3\2\2")
        buf.write("\29\67\3\2\2\2:<\7\6\2\2;:\3\2\2\2;<\3\2\2\2<>\3\2\2\2")
        buf.write("=\62\3\2\2\2=>\3\2\2\2>?\3\2\2\2?@\7\17\2\2@`\3\2\2\2")
        buf.write("AM\7\20\2\2BG\5\n\6\2CD\7\6\2\2DF\5\n\6\2EC\3\2\2\2FI")
        buf.write("\3\2\2\2GE\3\2\2\2GH\3\2\2\2HK\3\2\2\2IG\3\2\2\2JL\7\6")
        buf.write("\2\2KJ\3\2\2\2KL\3\2\2\2LN\3\2\2\2MB\3\2\2\2MN\3\2\2\2")
        buf.write("NO\3\2\2\2O`\7\21\2\2P\\\7\22\2\2QV\5\f\7\2RS\7\6\2\2")
        buf.write("SU\5\f\7\2TR\3\2\2\2UX\3\2\2\2VT\3\2\2\2VW\3\2\2\2WZ\3")
        buf.write("\2\2\2XV\3\2\2\2Y[\7\6\2\2ZY\3\2\2\2Z[\3\2\2\2[]\3\2\2")
        buf.write("\2\\Q\3\2\2\2\\]\3\2\2\2]^\3\2\2\2^`\7\23\2\2_/\3\2\2")
        buf.write("\2_\60\3\2\2\2_A\3\2\2\2_P\3\2\2\2`\7\3\2\2\2ab\5\24\13")
        buf.write("\2bc\7\f\2\2cd\5\16\b\2dl\3\2\2\2ef\5\24\13\2fg\7\f\2")
        buf.write("\2gh\7\20\2\2hi\5\16\b\2ij\7\21\2\2jl\3\2\2\2ka\3\2\2")
        buf.write("\2ke\3\2\2\2l\t\3\2\2\2ms\5\16\b\2no\7\20\2\2op\5\16\b")
        buf.write("\2pq\7\21\2\2qs\3\2\2\2rm\3\2\2\2rn\3\2\2\2s\13\3\2\2")
        buf.write("\2tu\5\24\13\2uv\7\b\2\2vw\5\16\b\2w\177\3\2\2\2xy\5\24")
        buf.write("\13\2yz\7\b\2\2z{\7\20\2\2{|\5\16\b\2|}\7\21\2\2}\177")
        buf.write("\3\2\2\2~t\3\2\2\2~x\3\2\2\2\177\r\3\2\2\2\u0080\u0094")
        buf.write("\7\7\2\2\u0081\u0094\5\26\f\2\u0082\u0083\7\16\2\2\u0083")
        buf.write("\u0086\5\26\f\2\u0084\u0085\7\r\2\2\u0085\u0087\5\26\f")
        buf.write("\2\u0086\u0084\3\2\2\2\u0087\u0088\3\2\2\2\u0088\u0086")
        buf.write("\3\2\2\2\u0088\u0089\3\2\2\2\u0089\u008a\3\2\2\2\u008a")
        buf.write("\u008b\7\17\2\2\u008b\u0094\3\2\2\2\u008c\u008f\5\26\f")
        buf.write("\2\u008d\u008e\7\r\2\2\u008e\u0090\5\26\f\2\u008f\u008d")
        buf.write("\3\2\2\2\u0090\u0091\3\2\2\2\u0091\u008f\3\2\2\2\u0091")
        buf.write("\u0092\3\2\2\2\u0092\u0094\3\2\2\2\u0093\u0080\3\2\2\2")
        buf.write("\u0093\u0081\3\2\2\2\u0093\u0082\3\2\2\2\u0093\u008c\3")
        buf.write("\2\2\2\u0094\u00ab\3\2\2\2\u0095\u00a2\7\7\2\2\u0096\u00a2")
        buf.write("\5\26\f\2\u0097\u0098\7\16\2\2\u0098\u009b\5\26\f\2\u0099")
        buf.write("\u009a\7\r\2\2\u009a\u009c\5\26\f\2\u009b\u0099\3\2\2")
        buf.write("\2\u009c\u009d\3\2\2\2\u009d\u009b\3\2\2\2\u009d\u009e")
        buf.write("\3\2\2\2\u009e\u009f\3\2\2\2\u009f\u00a0\7\17\2\2\u00a0")
        buf.write("\u00a2\3\2\2\2\u00a1\u0095\3\2\2\2\u00a1\u0096\3\2\2\2")
        buf.write("\u00a1\u0097\3\2\2\2\u00a2\u00a8\3\2\2\2\u00a3\u00a5\7")
        buf.write("\t\2\2\u00a4\u00a6\7\n\2\2\u00a5\u00a4\3\2\2\2\u00a5\u00a6")
        buf.write("\3\2\2\2\u00a6\u00a9\3\2\2\2\u00a7\u00a9\7\n\2\2\u00a8")
        buf.write("\u00a3\3\2\2\2\u00a8\u00a7\3\2\2\2\u00a9\u00ab\3\2\2\2")
        buf.write("\u00aa\u0093\3\2\2\2\u00aa\u00a1\3\2\2\2\u00ab\u00b8\3")
        buf.write("\2\2\2\u00ac\u00ae\5\30\r\2\u00ad\u00af\7\t\2\2\u00ae")
        buf.write("\u00ad\3\2\2\2\u00ae\u00af\3\2\2\2\u00af\u00b1\3\2\2\2")
        buf.write("\u00b0\u00b2\7\n\2\2\u00b1\u00b0\3\2\2\2\u00b1\u00b2\3")
        buf.write("\2\2\2\u00b2\u00b8\3\2\2\2\u00b3\u00b5\5\32\16\2\u00b4")
        buf.write("\u00b6\7\n\2\2\u00b5\u00b4\3\2\2\2\u00b5\u00b6\3\2\2\2")
        buf.write("\u00b6\u00b8\3\2\2\2\u00b7\u00aa\3\2\2\2\u00b7\u00ac\3")
        buf.write("\2\2\2\u00b7\u00b3\3\2\2\2\u00b8\17\3\2\2\2\u00b9\u00ba")
        buf.write("\t\2\2\2\u00ba\21\3\2\2\2\u00bb\u00bc\7\25\2\2\u00bc\23")
        buf.write("\3\2\2\2\u00bd\u00c0\5\20\t\2\u00be\u00c0\5\22\n\2\u00bf")
        buf.write("\u00bd\3\2\2\2\u00bf\u00be\3\2\2\2\u00c0\25\3\2\2\2\u00c1")
        buf.write("\u00c2\5\20\t\2\u00c2\27\3\2\2\2\u00c3\u00c4\5\22\n\2")
        buf.write("\u00c4\31\3\2\2\2\u00c5\u00c6\7\26\2\2\u00c6\33\3\2\2")
        buf.write("\2\35%\67;=GKMVZ\\_kr~\u0088\u0091\u0093\u009d\u00a1\u00a5")
        buf.write("\u00a8\u00aa\u00ae\u00b1\u00b5\u00b7\u00bf")
        return buf.getvalue()


class G4VisitorParser ( Parser ):

    grammarFileName = "G4Visitor.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'visitor'", "'for'", "'grammar'", "','", 
                     "'.'", "':'", "'*'", "'?'", "';'", "'='", "'|'", "'('", 
                     "')'", "'['", "']'", "'{'", "'}'" ]

    symbolicNames = [ "<INVALID>", "VISITOR", "FOR", "GRAMMAR", "COMMA", 
                      "DOT", "COLON", "STAR", "QUESTION", "SEMI", "EQUALS", 
                      "OR", "LPAREN", "RPAREN", "LBRACKET", "RBRACKET", 
                      "LBRACE", "RBRACE", "LC_NAME", "UC_NAME", "ATTRIBUTE_REF", 
                      "NUMBER", "COMMENT", "WS" ]

    RULE_visitor = 0
    RULE_visitorRule = 1
    RULE_body = 2
    RULE_objectParam = 3
    RULE_listItem = 4
    RULE_dictItem = 5
    RULE_expr = 6
    RULE_lcName = 7
    RULE_ucName = 8
    RULE_identifier = 9
    RULE_ruleName = 10
    RULE_tokenName = 11
    RULE_attributeRef = 12

    ruleNames =  [ "visitor", "visitorRule", "body", "objectParam", "listItem", 
                   "dictItem", "expr", "lcName", "ucName", "identifier", 
                   "ruleName", "tokenName", "attributeRef" ]

    EOF = Token.EOF
    VISITOR=1
    FOR=2
    GRAMMAR=3
    COMMA=4
    DOT=5
    COLON=6
    STAR=7
    QUESTION=8
    SEMI=9
    EQUALS=10
    OR=11
    LPAREN=12
    RPAREN=13
    LBRACKET=14
    RBRACKET=15
    LBRACE=16
    RBRACE=17
    LC_NAME=18
    UC_NAME=19
    ATTRIBUTE_REF=20
    NUMBER=21
    COMMENT=22
    WS=23

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class VisitorContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser
            self.name = None # IdentifierContext
            self.grammarName = None # IdentifierContext

        def VISITOR(self):
            return self.getToken(G4VisitorParser.VISITOR, 0)

        def FOR(self):
            return self.getToken(G4VisitorParser.FOR, 0)

        def GRAMMAR(self):
            return self.getToken(G4VisitorParser.GRAMMAR, 0)

        def SEMI(self):
            return self.getToken(G4VisitorParser.SEMI, 0)

        def EOF(self):
            return self.getToken(G4VisitorParser.EOF, 0)

        def identifier(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.IdentifierContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.IdentifierContext,i)


        def visitorRule(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.VisitorRuleContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.VisitorRuleContext,i)


        def getRuleIndex(self):
            return G4VisitorParser.RULE_visitor

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVisitor" ):
                return visitor.visitVisitor(self)
            else:
                return visitor.visitChildren(self)




    def visitor(self):

        localctx = G4VisitorParser.VisitorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_visitor)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 26
            self.match(G4VisitorParser.VISITOR)
            self.state = 27
            localctx.name = self.identifier()
            self.state = 28
            self.match(G4VisitorParser.FOR)
            self.state = 29
            self.match(G4VisitorParser.GRAMMAR)
            self.state = 30
            localctx.grammarName = self.identifier()
            self.state = 31
            self.match(G4VisitorParser.SEMI)
            self.state = 35
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << G4VisitorParser.VISITOR) | (1 << G4VisitorParser.FOR) | (1 << G4VisitorParser.GRAMMAR) | (1 << G4VisitorParser.LC_NAME))) != 0):
                self.state = 32
                self.visitorRule()
                self.state = 37
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 38
            self.match(G4VisitorParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class VisitorRuleContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ruleName(self):
            return self.getTypedRuleContext(G4VisitorParser.RuleNameContext,0)


        def EQUALS(self):
            return self.getToken(G4VisitorParser.EQUALS, 0)

        def body(self):
            return self.getTypedRuleContext(G4VisitorParser.BodyContext,0)


        def SEMI(self):
            return self.getToken(G4VisitorParser.SEMI, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_visitorRule

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitVisitorRule" ):
                return visitor.visitVisitorRule(self)
            else:
                return visitor.visitChildren(self)




    def visitorRule(self):

        localctx = G4VisitorParser.VisitorRuleContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_visitorRule)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 40
            self.ruleName()
            self.state = 41
            self.match(G4VisitorParser.EQUALS)
            self.state = 42
            self.body()
            self.state = 43
            self.match(G4VisitorParser.SEMI)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class BodyContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return G4VisitorParser.RULE_body

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class ExprBodyContext(BodyContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.BodyContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(G4VisitorParser.ExprContext,0)


        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExprBody" ):
                return visitor.visitExprBody(self)
            else:
                return visitor.visitChildren(self)


    class ListBodyContext(BodyContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.BodyContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACKET(self):
            return self.getToken(G4VisitorParser.LBRACKET, 0)
        def RBRACKET(self):
            return self.getToken(G4VisitorParser.RBRACKET, 0)
        def listItem(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.ListItemContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.ListItemContext,i)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(G4VisitorParser.COMMA)
            else:
                return self.getToken(G4VisitorParser.COMMA, i)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListBody" ):
                return visitor.visitListBody(self)
            else:
                return visitor.visitChildren(self)


    class DictBodyContext(BodyContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.BodyContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LBRACE(self):
            return self.getToken(G4VisitorParser.LBRACE, 0)
        def RBRACE(self):
            return self.getToken(G4VisitorParser.RBRACE, 0)
        def dictItem(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.DictItemContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.DictItemContext,i)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(G4VisitorParser.COMMA)
            else:
                return self.getToken(G4VisitorParser.COMMA, i)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDictBody" ):
                return visitor.visitDictBody(self)
            else:
                return visitor.visitChildren(self)


    class ObjectBodyContext(BodyContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.BodyContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def identifier(self):
            return self.getTypedRuleContext(G4VisitorParser.IdentifierContext,0)

        def LPAREN(self):
            return self.getToken(G4VisitorParser.LPAREN, 0)
        def RPAREN(self):
            return self.getToken(G4VisitorParser.RPAREN, 0)
        def objectParam(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.ObjectParamContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.ObjectParamContext,i)

        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(G4VisitorParser.COMMA)
            else:
                return self.getToken(G4VisitorParser.COMMA, i)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectBody" ):
                return visitor.visitObjectBody(self)
            else:
                return visitor.visitChildren(self)



    def body(self):

        localctx = G4VisitorParser.BodyContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_body)
        self._la = 0 # Token type
        try:
            self.state = 93
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,10,self._ctx)
            if la_ == 1:
                localctx = G4VisitorParser.ExprBodyContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 45
                self.expr()
                pass

            elif la_ == 2:
                localctx = G4VisitorParser.ObjectBodyContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 46
                self.identifier()
                self.state = 47
                self.match(G4VisitorParser.LPAREN)
                self.state = 59
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << G4VisitorParser.VISITOR) | (1 << G4VisitorParser.FOR) | (1 << G4VisitorParser.GRAMMAR) | (1 << G4VisitorParser.LC_NAME) | (1 << G4VisitorParser.UC_NAME))) != 0):
                    self.state = 48
                    self.objectParam()
                    self.state = 53
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,1,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 49
                            self.match(G4VisitorParser.COMMA)
                            self.state = 50
                            self.objectParam() 
                        self.state = 55
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,1,self._ctx)

                    self.state = 57
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==G4VisitorParser.COMMA:
                        self.state = 56
                        self.match(G4VisitorParser.COMMA)




                self.state = 61
                self.match(G4VisitorParser.RPAREN)
                pass

            elif la_ == 3:
                localctx = G4VisitorParser.ListBodyContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 63
                self.match(G4VisitorParser.LBRACKET)
                self.state = 75
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << G4VisitorParser.VISITOR) | (1 << G4VisitorParser.FOR) | (1 << G4VisitorParser.GRAMMAR) | (1 << G4VisitorParser.DOT) | (1 << G4VisitorParser.LPAREN) | (1 << G4VisitorParser.LBRACKET) | (1 << G4VisitorParser.LC_NAME) | (1 << G4VisitorParser.UC_NAME) | (1 << G4VisitorParser.ATTRIBUTE_REF))) != 0):
                    self.state = 64
                    self.listItem()
                    self.state = 69
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,4,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 65
                            self.match(G4VisitorParser.COMMA)
                            self.state = 66
                            self.listItem() 
                        self.state = 71
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,4,self._ctx)

                    self.state = 73
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==G4VisitorParser.COMMA:
                        self.state = 72
                        self.match(G4VisitorParser.COMMA)




                self.state = 77
                self.match(G4VisitorParser.RBRACKET)
                pass

            elif la_ == 4:
                localctx = G4VisitorParser.DictBodyContext(self, localctx)
                self.enterOuterAlt(localctx, 4)
                self.state = 78
                self.match(G4VisitorParser.LBRACE)
                self.state = 90
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << G4VisitorParser.VISITOR) | (1 << G4VisitorParser.FOR) | (1 << G4VisitorParser.GRAMMAR) | (1 << G4VisitorParser.LC_NAME) | (1 << G4VisitorParser.UC_NAME))) != 0):
                    self.state = 79
                    self.dictItem()
                    self.state = 84
                    self._errHandler.sync(self)
                    _alt = self._interp.adaptivePredict(self._input,7,self._ctx)
                    while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                        if _alt==1:
                            self.state = 80
                            self.match(G4VisitorParser.COMMA)
                            self.state = 81
                            self.dictItem() 
                        self.state = 86
                        self._errHandler.sync(self)
                        _alt = self._interp.adaptivePredict(self._input,7,self._ctx)

                    self.state = 88
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if _la==G4VisitorParser.COMMA:
                        self.state = 87
                        self.match(G4VisitorParser.COMMA)




                self.state = 92
                self.match(G4VisitorParser.RBRACE)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ObjectParamContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self):
            return self.getTypedRuleContext(G4VisitorParser.IdentifierContext,0)


        def EQUALS(self):
            return self.getToken(G4VisitorParser.EQUALS, 0)

        def expr(self):
            return self.getTypedRuleContext(G4VisitorParser.ExprContext,0)


        def LBRACKET(self):
            return self.getToken(G4VisitorParser.LBRACKET, 0)

        def RBRACKET(self):
            return self.getToken(G4VisitorParser.RBRACKET, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_objectParam

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitObjectParam" ):
                return visitor.visitObjectParam(self)
            else:
                return visitor.visitChildren(self)




    def objectParam(self):

        localctx = G4VisitorParser.ObjectParamContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_objectParam)
        try:
            self.state = 105
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,11,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 95
                self.identifier()
                self.state = 96
                self.match(G4VisitorParser.EQUALS)
                self.state = 97
                self.expr()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 99
                self.identifier()
                self.state = 100
                self.match(G4VisitorParser.EQUALS)
                self.state = 101
                self.match(G4VisitorParser.LBRACKET)
                self.state = 102
                self.expr()
                self.state = 103
                self.match(G4VisitorParser.RBRACKET)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ListItemContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(G4VisitorParser.ExprContext,0)


        def LBRACKET(self):
            return self.getToken(G4VisitorParser.LBRACKET, 0)

        def RBRACKET(self):
            return self.getToken(G4VisitorParser.RBRACKET, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_listItem

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitListItem" ):
                return visitor.visitListItem(self)
            else:
                return visitor.visitChildren(self)




    def listItem(self):

        localctx = G4VisitorParser.ListItemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_listItem)
        try:
            self.state = 112
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [G4VisitorParser.VISITOR, G4VisitorParser.FOR, G4VisitorParser.GRAMMAR, G4VisitorParser.DOT, G4VisitorParser.LPAREN, G4VisitorParser.LC_NAME, G4VisitorParser.UC_NAME, G4VisitorParser.ATTRIBUTE_REF]:
                self.enterOuterAlt(localctx, 1)
                self.state = 107
                self.expr()
                pass
            elif token in [G4VisitorParser.LBRACKET]:
                self.enterOuterAlt(localctx, 2)
                self.state = 108
                self.match(G4VisitorParser.LBRACKET)
                self.state = 109
                self.expr()
                self.state = 110
                self.match(G4VisitorParser.RBRACKET)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DictItemContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def identifier(self):
            return self.getTypedRuleContext(G4VisitorParser.IdentifierContext,0)


        def COLON(self):
            return self.getToken(G4VisitorParser.COLON, 0)

        def expr(self):
            return self.getTypedRuleContext(G4VisitorParser.ExprContext,0)


        def LBRACKET(self):
            return self.getToken(G4VisitorParser.LBRACKET, 0)

        def RBRACKET(self):
            return self.getToken(G4VisitorParser.RBRACKET, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_dictItem

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitDictItem" ):
                return visitor.visitDictItem(self)
            else:
                return visitor.visitChildren(self)




    def dictItem(self):

        localctx = G4VisitorParser.DictItemContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_dictItem)
        try:
            self.state = 124
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,13,self._ctx)
            if la_ == 1:
                self.enterOuterAlt(localctx, 1)
                self.state = 114
                self.identifier()
                self.state = 115
                self.match(G4VisitorParser.COLON)
                self.state = 116
                self.expr()
                pass

            elif la_ == 2:
                self.enterOuterAlt(localctx, 2)
                self.state = 118
                self.identifier()
                self.state = 119
                self.match(G4VisitorParser.COLON)
                self.state = 120
                self.match(G4VisitorParser.LBRACKET)
                self.state = 121
                self.expr()
                self.state = 122
                self.match(G4VisitorParser.RBRACKET)
                pass


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return G4VisitorParser.RULE_expr

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class TokenExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def tokenName(self):
            return self.getTypedRuleContext(G4VisitorParser.TokenNameContext,0)

        def STAR(self):
            return self.getToken(G4VisitorParser.STAR, 0)
        def QUESTION(self):
            return self.getToken(G4VisitorParser.QUESTION, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTokenExpr" ):
                return visitor.visitTokenExpr(self)
            else:
                return visitor.visitChildren(self)


    class RefExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def attributeRef(self):
            return self.getTypedRuleContext(G4VisitorParser.AttributeRefContext,0)

        def QUESTION(self):
            return self.getToken(G4VisitorParser.QUESTION, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRefExpr" ):
                return visitor.visitRefExpr(self)
            else:
                return visitor.visitChildren(self)


    class RuleExprContext(ExprContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a G4VisitorParser.ExprContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def DOT(self):
            return self.getToken(G4VisitorParser.DOT, 0)
        def ruleName(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(G4VisitorParser.RuleNameContext)
            else:
                return self.getTypedRuleContext(G4VisitorParser.RuleNameContext,i)

        def LPAREN(self):
            return self.getToken(G4VisitorParser.LPAREN, 0)
        def RPAREN(self):
            return self.getToken(G4VisitorParser.RPAREN, 0)
        def STAR(self):
            return self.getToken(G4VisitorParser.STAR, 0)
        def QUESTION(self):
            return self.getToken(G4VisitorParser.QUESTION, 0)
        def OR(self, i:int=None):
            if i is None:
                return self.getTokens(G4VisitorParser.OR)
            else:
                return self.getToken(G4VisitorParser.OR, i)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRuleExpr" ):
                return visitor.visitRuleExpr(self)
            else:
                return visitor.visitChildren(self)



    def expr(self):

        localctx = G4VisitorParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_expr)
        self._la = 0 # Token type
        try:
            self.state = 181
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [G4VisitorParser.VISITOR, G4VisitorParser.FOR, G4VisitorParser.GRAMMAR, G4VisitorParser.DOT, G4VisitorParser.LPAREN, G4VisitorParser.LC_NAME]:
                localctx = G4VisitorParser.RuleExprContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 168
                self._errHandler.sync(self)
                la_ = self._interp.adaptivePredict(self._input,21,self._ctx)
                if la_ == 1:
                    self.state = 145
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,16,self._ctx)
                    if la_ == 1:
                        self.state = 126
                        self.match(G4VisitorParser.DOT)
                        pass

                    elif la_ == 2:
                        self.state = 127
                        self.ruleName()
                        pass

                    elif la_ == 3:
                        self.state = 128
                        self.match(G4VisitorParser.LPAREN)
                        self.state = 129
                        self.ruleName()
                        self.state = 132 
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        while True:
                            self.state = 130
                            self.match(G4VisitorParser.OR)
                            self.state = 131
                            self.ruleName()
                            self.state = 134 
                            self._errHandler.sync(self)
                            _la = self._input.LA(1)
                            if not (_la==G4VisitorParser.OR):
                                break

                        self.state = 136
                        self.match(G4VisitorParser.RPAREN)
                        pass

                    elif la_ == 4:
                        self.state = 138
                        self.ruleName()
                        self.state = 141 
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        while True:
                            self.state = 139
                            self.match(G4VisitorParser.OR)
                            self.state = 140
                            self.ruleName()
                            self.state = 143 
                            self._errHandler.sync(self)
                            _la = self._input.LA(1)
                            if not (_la==G4VisitorParser.OR):
                                break

                        pass


                    pass

                elif la_ == 2:
                    self.state = 159
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [G4VisitorParser.DOT]:
                        self.state = 147
                        self.match(G4VisitorParser.DOT)
                        pass
                    elif token in [G4VisitorParser.VISITOR, G4VisitorParser.FOR, G4VisitorParser.GRAMMAR, G4VisitorParser.LC_NAME]:
                        self.state = 148
                        self.ruleName()
                        pass
                    elif token in [G4VisitorParser.LPAREN]:
                        self.state = 149
                        self.match(G4VisitorParser.LPAREN)
                        self.state = 150
                        self.ruleName()
                        self.state = 153 
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        while True:
                            self.state = 151
                            self.match(G4VisitorParser.OR)
                            self.state = 152
                            self.ruleName()
                            self.state = 155 
                            self._errHandler.sync(self)
                            _la = self._input.LA(1)
                            if not (_la==G4VisitorParser.OR):
                                break

                        self.state = 157
                        self.match(G4VisitorParser.RPAREN)
                        pass
                    else:
                        raise NoViableAltException(self)

                    self.state = 166
                    self._errHandler.sync(self)
                    token = self._input.LA(1)
                    if token in [G4VisitorParser.STAR]:
                        self.state = 161
                        self.match(G4VisitorParser.STAR)
                        self.state = 163
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)
                        if _la==G4VisitorParser.QUESTION:
                            self.state = 162
                            self.match(G4VisitorParser.QUESTION)


                        pass
                    elif token in [G4VisitorParser.QUESTION]:
                        self.state = 165
                        self.match(G4VisitorParser.QUESTION)
                        pass
                    else:
                        raise NoViableAltException(self)

                    pass


                pass
            elif token in [G4VisitorParser.UC_NAME]:
                localctx = G4VisitorParser.TokenExprContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 170
                self.tokenName()
                self.state = 172
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==G4VisitorParser.STAR:
                    self.state = 171
                    self.match(G4VisitorParser.STAR)


                self.state = 175
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==G4VisitorParser.QUESTION:
                    self.state = 174
                    self.match(G4VisitorParser.QUESTION)


                pass
            elif token in [G4VisitorParser.ATTRIBUTE_REF]:
                localctx = G4VisitorParser.RefExprContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 177
                self.attributeRef()
                self.state = 179
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==G4VisitorParser.QUESTION:
                    self.state = 178
                    self.match(G4VisitorParser.QUESTION)


                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LcNameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LC_NAME(self):
            return self.getToken(G4VisitorParser.LC_NAME, 0)

        def VISITOR(self):
            return self.getToken(G4VisitorParser.VISITOR, 0)

        def FOR(self):
            return self.getToken(G4VisitorParser.FOR, 0)

        def GRAMMAR(self):
            return self.getToken(G4VisitorParser.GRAMMAR, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_lcName

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitLcName" ):
                return visitor.visitLcName(self)
            else:
                return visitor.visitChildren(self)




    def lcName(self):

        localctx = G4VisitorParser.LcNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_lcName)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 183
            _la = self._input.LA(1)
            if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << G4VisitorParser.VISITOR) | (1 << G4VisitorParser.FOR) | (1 << G4VisitorParser.GRAMMAR) | (1 << G4VisitorParser.LC_NAME))) != 0)):
                self._errHandler.recoverInline(self)
            else:
                self._errHandler.reportMatch(self)
                self.consume()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class UcNameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def UC_NAME(self):
            return self.getToken(G4VisitorParser.UC_NAME, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_ucName

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitUcName" ):
                return visitor.visitUcName(self)
            else:
                return visitor.visitChildren(self)




    def ucName(self):

        localctx = G4VisitorParser.UcNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_ucName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 185
            self.match(G4VisitorParser.UC_NAME)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class IdentifierContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lcName(self):
            return self.getTypedRuleContext(G4VisitorParser.LcNameContext,0)


        def ucName(self):
            return self.getTypedRuleContext(G4VisitorParser.UcNameContext,0)


        def getRuleIndex(self):
            return G4VisitorParser.RULE_identifier

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIdentifier" ):
                return visitor.visitIdentifier(self)
            else:
                return visitor.visitChildren(self)




    def identifier(self):

        localctx = G4VisitorParser.IdentifierContext(self, self._ctx, self.state)
        self.enterRule(localctx, 18, self.RULE_identifier)
        try:
            self.state = 189
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [G4VisitorParser.VISITOR, G4VisitorParser.FOR, G4VisitorParser.GRAMMAR, G4VisitorParser.LC_NAME]:
                self.enterOuterAlt(localctx, 1)
                self.state = 187
                self.lcName()
                pass
            elif token in [G4VisitorParser.UC_NAME]:
                self.enterOuterAlt(localctx, 2)
                self.state = 188
                self.ucName()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RuleNameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def lcName(self):
            return self.getTypedRuleContext(G4VisitorParser.LcNameContext,0)


        def getRuleIndex(self):
            return G4VisitorParser.RULE_ruleName

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitRuleName" ):
                return visitor.visitRuleName(self)
            else:
                return visitor.visitChildren(self)




    def ruleName(self):

        localctx = G4VisitorParser.RuleNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_ruleName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 191
            self.lcName()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TokenNameContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ucName(self):
            return self.getTypedRuleContext(G4VisitorParser.UcNameContext,0)


        def getRuleIndex(self):
            return G4VisitorParser.RULE_tokenName

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitTokenName" ):
                return visitor.visitTokenName(self)
            else:
                return visitor.visitChildren(self)




    def tokenName(self):

        localctx = G4VisitorParser.TokenNameContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_tokenName)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 193
            self.ucName()
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AttributeRefContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def ATTRIBUTE_REF(self):
            return self.getToken(G4VisitorParser.ATTRIBUTE_REF, 0)

        def getRuleIndex(self):
            return G4VisitorParser.RULE_attributeRef

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitAttributeRef" ):
                return visitor.visitAttributeRef(self)
            else:
                return visitor.visitChildren(self)




    def attributeRef(self):

        localctx = G4VisitorParser.AttributeRefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_attributeRef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 195
            self.match(G4VisitorParser.ATTRIBUTE_REF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








from ruamel.yaml import YAML as _YAML
from .dot_dict import DotDict


class DotDictConstructionMixin(object):
    def construct_yaml_map(self, node):
        data = DotDict()
        yield data
        value = self.construct_mapping(node)
        data.update(value)


def YAML(**kwargs):
    yaml = _YAML(**kwargs)

    class Constructor(DotDictConstructionMixin, yaml.Constructor):
        pass

    Constructor.add_constructor(
        u'tag:yaml.org,2002:map',
        Constructor.construct_yaml_map)

    yaml.Constructor = Constructor
    yaml.Representer.add_representer(DotDict, yaml.Representer.yaml_representers[dict])

    return yaml


from antlr4 import InputStream, FileStream, CommonTokenStream
from antlr4.error.Errors import ParseCancellationException
from antlr4.error.ErrorListener import ErrorListener
from .dot_dict import DotDict


class ParseTreeVisitor(object):
    # mandatory visitor methods

    def visitChildren(self, node, *types):
        return self.visitNodes(self.get_children(node, *types))

    def visitTerminal(self, node):
        return node.symbol.text

    def visitErrorNode(self, node):
        return None

    # convenience visitor methods

    def visitChild(self, node, *types):
        return self.visitNode(self.get_child(node, *types))

    def visitNodes(self, nodes):
        return [node.accept(self) for node in nodes]

    def visitNode(self, node):
        return node.accept(self)

    # auxillary methods

    def get_children(self, node, *types):
        return (child for child in node.getChildren() if len(types) == 0 or isinstance(child, types))

    def has_children(self, node, *types):
        it = self.get_children(node, *types)
        try:
            next(it)
        except StopIteration:
            return False
        else:
            return True

    def get_child(self, node, *types):
        child, = self.get_children(node, *types)
        return child

    def collapse(self, nodes, *, constructor=None, name=None):
        if constructor is not None and name is not None:
            raise ValueError("constructor and name are mutually exclusive")
        try:
            node, = nodes
        except ValueError:
            if constructor is not None:
                return constructor(nodes)
            elif name is not None:
                return {name: nodes}
            else:
                return nodes
        else:
            return node

    def get_text(self, node):
        return node.getText()

    def get_full_text(self, node):
        return node.parser.getTokenStream().getText(node.getSourceInterval())


class DotDictVisitorMixin(object):
    def visitChildren(self, node):
        result = DotDict()
        result[self.rule_name(node)] = [child.accept(self) for child in node.getChildren()]
        return result

    def aggregateResult(self, aggregate, nextResult):
        aggregate[nextResult] = None
        return aggregate

    def visitTerminal(self, node):
        return self.token_name(node.symbol)

    def rule_name(self, node):
        return self.Parser.ruleNames[node.getRuleIndex()]

    def token_name(self, token):
        return token.text

    # convenience methods

    def visitSingleChild(self, node):
        assert node.getChildCount() == 1, "visitSingleChild() requires there to be only one child"
        return node.getChild(0).accept(self)


def dot_dict_visitor(_Parser, _Visitor):
    class Visitor(DotDictVisitorMixin, _Visitor):
        Parser = _Parser

    return Visitor


class BailErrorListener(ErrorListener):
    INSTANCE = None

    def syntaxError(self, recognizer, offendingSymbol, line, column, msg, e):
        raise ParseCancellationException(f"line {line}:{column} {msg}")


BailErrorListener.INSTANCE = BailErrorListener()


class Antlr(object):
    def __init__(self, Lexer=None, Parser=None):
        self.Lexer = Lexer
        self.Parser = Parser

    def input_stream(self, text):
        return InputStream(text)

    def file_stream(self, file, encoding='ascii', errors='strict'):
        return FileStream(file, encoding, errors)

    def lexer(self, input):
        lexer = self.Lexer(input)
        lexer.removeErrorListeners()
        return lexer

    def token_stream(self, input):
        return CommonTokenStream(self.lexer(input))

    def parser(self, input):
        parser = self.Parser(self.token_stream(input))
        parser.removeErrorListeners()
        parser.addErrorListener(BailErrorListener.INSTANCE)
        return parser

    def visitor(self, Visitor):
        return dot_dict_visitor(self.Parser, Visitor)

    def to_string(self, ctx):
        return ctx.toStringTree(self.Parser.ruleNames, self.Parser)

    def parse_safe(self, parse):
        try:
            return parse()
        except ParseCancellationException:
            return None



# Generated from gsl/grammar/G4Visitor.g4 by ANTLR 4.7
from antlr4 import *
if __name__ is not None and "." in __name__:
    from .G4VisitorParser import G4VisitorParser
else:
    from G4VisitorParser import G4VisitorParser

# This class defines a complete generic visitor for a parse tree produced by G4VisitorParser.

class G4VisitorVisitor(ParseTreeVisitor):

    # Visit a parse tree produced by G4VisitorParser#visitor.
    def visitVisitor(self, ctx:G4VisitorParser.VisitorContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#visitorRule.
    def visitVisitorRule(self, ctx:G4VisitorParser.VisitorRuleContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#exprBody.
    def visitExprBody(self, ctx:G4VisitorParser.ExprBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#objectBody.
    def visitObjectBody(self, ctx:G4VisitorParser.ObjectBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#listBody.
    def visitListBody(self, ctx:G4VisitorParser.ListBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#dictBody.
    def visitDictBody(self, ctx:G4VisitorParser.DictBodyContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#objectParam.
    def visitObjectParam(self, ctx:G4VisitorParser.ObjectParamContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#listItem.
    def visitListItem(self, ctx:G4VisitorParser.ListItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#dictItem.
    def visitDictItem(self, ctx:G4VisitorParser.DictItemContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#ruleExpr.
    def visitRuleExpr(self, ctx:G4VisitorParser.RuleExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#tokenExpr.
    def visitTokenExpr(self, ctx:G4VisitorParser.TokenExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#refExpr.
    def visitRefExpr(self, ctx:G4VisitorParser.RefExprContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#lcName.
    def visitLcName(self, ctx:G4VisitorParser.LcNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#ucName.
    def visitUcName(self, ctx:G4VisitorParser.UcNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#identifier.
    def visitIdentifier(self, ctx:G4VisitorParser.IdentifierContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#ruleName.
    def visitRuleName(self, ctx:G4VisitorParser.RuleNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#tokenName.
    def visitTokenName(self, ctx:G4VisitorParser.TokenNameContext):
        return self.visitChildren(ctx)


    # Visit a parse tree produced by G4VisitorParser#attributeRef.
    def visitAttributeRef(self, ctx:G4VisitorParser.AttributeRefContext):
        return self.visitChildren(ctx)



del G4VisitorParser


# Generated from tests/grammar/SetTest.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\7")
        buf.write("!\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\3\2\3\2\3\2\3\3\3\3")
        buf.write("\3\3\3\3\7\3\22\n\3\f\3\16\3\25\13\3\5\3\27\n\3\3\3\3")
        buf.write("\3\3\4\3\4\5\4\35\n\4\3\5\3\5\3\5\2\2\6\2\4\6\b\2\2\2")
        buf.write("\37\2\n\3\2\2\2\4\r\3\2\2\2\6\34\3\2\2\2\b\36\3\2\2\2")
        buf.write("\n\13\5\4\3\2\13\f\7\2\2\3\f\3\3\2\2\2\r\26\7\4\2\2\16")
        buf.write("\23\5\6\4\2\17\20\7\3\2\2\20\22\5\6\4\2\21\17\3\2\2\2")
        buf.write("\22\25\3\2\2\2\23\21\3\2\2\2\23\24\3\2\2\2\24\27\3\2\2")
        buf.write("\2\25\23\3\2\2\2\26\16\3\2\2\2\26\27\3\2\2\2\27\30\3\2")
        buf.write("\2\2\30\31\7\5\2\2\31\5\3\2\2\2\32\35\5\b\5\2\33\35\5")
        buf.write("\4\3\2\34\32\3\2\2\2\34\33\3\2\2\2\35\7\3\2\2\2\36\37")
        buf.write("\7\6\2\2\37\t\3\2\2\2\5\23\26\34")
        return buf.getvalue()


class SetTestParser ( Parser ):

    grammarFileName = "SetTest.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "','", "'{'", "'}'" ]

    symbolicNames = [ "<INVALID>", "COMMA", "LBRACE", "RBRACE", "NUMBER", 
                      "WS" ]

    RULE_expr = 0
    RULE_set0 = 1
    RULE_element = 2
    RULE_simpleElement = 3

    ruleNames =  [ "expr", "set0", "element", "simpleElement" ]

    EOF = Token.EOF
    COMMA=1
    LBRACE=2
    RBRACE=3
    NUMBER=4
    WS=5

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def set0(self):
            return self.getTypedRuleContext(SetTestParser.Set0Context,0)


        def EOF(self):
            return self.getToken(SetTestParser.EOF, 0)

        def getRuleIndex(self):
            return SetTestParser.RULE_expr

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpr" ):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)




    def expr(self):

        localctx = SetTestParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_expr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.set0()
            self.state = 9
            self.match(SetTestParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Set0Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def LBRACE(self):
            return self.getToken(SetTestParser.LBRACE, 0)

        def RBRACE(self):
            return self.getToken(SetTestParser.RBRACE, 0)

        def element(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(SetTestParser.ElementContext)
            else:
                return self.getTypedRuleContext(SetTestParser.ElementContext,i)


        def COMMA(self, i:int=None):
            if i is None:
                return self.getTokens(SetTestParser.COMMA)
            else:
                return self.getToken(SetTestParser.COMMA, i)

        def getRuleIndex(self):
            return SetTestParser.RULE_set0

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSet0" ):
                return visitor.visitSet0(self)
            else:
                return visitor.visitChildren(self)




    def set0(self):

        localctx = SetTestParser.Set0Context(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_set0)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 11
            self.match(SetTestParser.LBRACE)
            self.state = 20
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==SetTestParser.LBRACE or _la==SetTestParser.NUMBER:
                self.state = 12
                self.element()
                self.state = 17
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==SetTestParser.COMMA:
                    self.state = 13
                    self.match(SetTestParser.COMMA)
                    self.state = 14
                    self.element()
                    self.state = 19
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 22
            self.match(SetTestParser.RBRACE)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ElementContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def simpleElement(self):
            return self.getTypedRuleContext(SetTestParser.SimpleElementContext,0)


        def set0(self):
            return self.getTypedRuleContext(SetTestParser.Set0Context,0)


        def getRuleIndex(self):
            return SetTestParser.RULE_element

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitElement" ):
                return visitor.visitElement(self)
            else:
                return visitor.visitChildren(self)




    def element(self):

        localctx = SetTestParser.ElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_element)
        try:
            self.state = 26
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [SetTestParser.NUMBER]:
                self.enterOuterAlt(localctx, 1)
                self.state = 24
                self.simpleElement()
                pass
            elif token in [SetTestParser.LBRACE]:
                self.enterOuterAlt(localctx, 2)
                self.state = 25
                self.set0()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SimpleElementContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return SetTestParser.RULE_simpleElement

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class IntElementContext(SimpleElementContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a SetTestParser.SimpleElementContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUMBER(self):
            return self.getToken(SetTestParser.NUMBER, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitIntElement" ):
                return visitor.visitIntElement(self)
            else:
                return visitor.visitChildren(self)



    def simpleElement(self):

        localctx = SetTestParser.SimpleElementContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_simpleElement)
        try:
            localctx = SetTestParser.IntElementContext(self, localctx)
            self.enterOuterAlt(localctx, 1)
            self.state = 28
            self.match(SetTestParser.NUMBER)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx








import re


def case(*, to, **kwargs):
    """Converts an identifier from one case type to another.
    An identifier is an ASCII string consisting of letters, digits and underscores, not starting with a digit.
    The supported case types are camelCase, PascalCase, snake_case, and CONSTANT_CASE,
    identified as camel, pascal, snake, and constant.
    The input identifier is given as a keyword argument with one of these names,
    and the output type is given as a string in the `to` keyword argument.
    If a given string does not conform to the specified case type (such as underscores in camel or pascal case strings,
    or double__underscores in general), the result may not be as desired,
    although things like snaKe_casE or CONStaNT_CASe will generally work."""

    if len(kwargs) != 1:
        raise ValueError("expect exactly one source string argument")

    [(typ, string)] = kwargs.items()

    types = {'pascal', 'camel', 'snake', 'constant'}
    if typ not in types:
        raise ValueError(f"source string keyword must be one of {types}")
    if to not in types:
        raise ValueError(f"\"to\" argument must be one of {types}")

    def pascal_iter(string):
        yield from (m.group(0) for m in re.finditer(r'[A-Z][a-z0-9]*|[a-z0-9]+', string))

    def snake_iter(string):
        yield from (m.group(2) for m in re.finditer(r'(^|_)([A-Za-z0-9]+)', string))

    inputs = {
        'pascal': pascal_iter,
        'camel': pascal_iter,
        'snake': snake_iter,
        'constant': snake_iter,
    }

    def out_fun(sep, case=None, case_fst=None):
        if case is None:
            case = lambda x: x
        if case_fst is None:
            case_fst = case
        return lambda tokens: sep.join(case_fst(token) if i == 0 else case(token) for i, token in enumerate(tokens))

    outputs = {
        'pascal': out_fun('', str.capitalize),
        'camel': out_fun('', str.capitalize, str.lower),
        'snake': out_fun('_', str.lower),
        'constant': out_fun('_', str.upper),
    }

    tokens = inputs[typ](string)
    return outputs[to](tokens)



from collections import namedtuple

from gsl.antlr import ParseTreeVisitor
from .ExprTestParser import ExprTestParser




class ExprTestVisitor(ParseTreeVisitor):
    def visitExpr(self, ctx: ExprTestParser.ExprContext):
        return self.visitNode(self.get_child(ctx, ExprTestParser.Expr0Context))

    def visitExpr0(self, ctx: ExprTestParser.Expr0Context):
        return self.visitNodes(self.get_children(ctx))

    def visitSummand(self, ctx: ExprTestParser.SummandContext):
        return self.visitNodes(self.get_children(ctx))

    def visitNumber(self, ctx: ExprTestParser.NumberContext):
        return self.visitNode(self.get_child(ctx))

    def visitSubexpr(self, ctx: ExprTestParser.SubexprContext):
        return self.visitNode(self.get_child(ctx, ExprTestParser.Expr0Context))




from gsl import pseudo_tuple

from gsl.antlr import ParseTreeVisitor
if __name__ is not None and "." in __name__:
    from .G4VisitorParser import G4VisitorParser
else:
    from G4VisitorParser import G4VisitorParser


Visitor = pseudo_tuple('Visitor', ('name', 'grammar', 'rules',))
Rule = pseudo_tuple('Rule', ('name', 'body',))
ObjectBody = pseudo_tuple('ObjectBody', ('name', 'params',))
ObjectParam = pseudo_tuple('ObjectParam', ('name', 'expr', 'optional',))
ListItem = pseudo_tuple('ListItem', ('expr', 'optional',))
DictItem = pseudo_tuple('DictItem', ('name', 'expr', 'optional',))
RuleExpr = pseudo_tuple('RuleExpr', ('rules', 'multi', 'presence',))
TokenExpr = pseudo_tuple('TokenExpr', ('token', 'multi', 'presence',))
RefExpr = pseudo_tuple('RefExpr', ('ref', 'presence',))


class G4VVisitor(ParseTreeVisitor):
    def visitVisitor(self, ctx: G4VisitorParser.VisitorContext):
        return Visitor(
            self.visitNode(ctx.name),
            self.visitNode(ctx.grammarName),
            self.visitNodes(self.get_children(ctx, G4VisitorParser.VisitorRuleContext)),
        )

    def visitVisitorRule(self, ctx: G4VisitorParser.VisitorRuleContext):
        return Rule(
            self.visitNode(self.get_child(ctx, G4VisitorParser.RuleNameContext)),
            self.visitNode(self.get_child(ctx, G4VisitorParser.BodyContext)),
        )

    def visitExprBody(self, ctx: G4VisitorParser.ExprBodyContext):
        return self.visitNode(self.get_child(ctx, G4VisitorParser.ExprContext))

    def visitObjectBody(self, ctx: G4VisitorParser.ObjectBodyContext):
        return ObjectBody(
            self.visitNode(self.get_child(ctx, G4VisitorParser.IdentifierContext)),
            self.visitNodes(self.get_children(ctx, G4VisitorParser.ObjectParamContext)),
        )

    def visitListBody(self, ctx: G4VisitorParser.ListBodyContext):
        return self.visitNodes(self.get_children(ctx, G4VisitorParser.ListItemContext))

    def visitDictBody(self, ctx: G4VisitorParser.DictBodyContext):
        return self.visitNodes(self.get_children(ctx, G4VisitorParser.DictItemContext))

    def visitObjectParam(self, ctx: G4VisitorParser.ObjectParamContext):
        return ObjectParam(
            self.visitNode(self.get_child(ctx, G4VisitorParser.IdentifierContext)),
            self.visitNode(self.get_child(ctx, G4VisitorParser.ExprContext)),
            bool(ctx.LBRACKET()),
        )

    def visitListItem(self, ctx: G4VisitorParser.ListItemContext):
        return ListItem(
            self.visitNode(self.get_child(ctx, G4VisitorParser.ExprContext)),
            bool(ctx.LBRACKET()),
        )

    def visitDictItem(self, ctx: G4VisitorParser.DictItemContext):
        return DictItem(
            self.visitNode(self.get_child(ctx, G4VisitorParser.IdentifierContext)),
            self.visitNode(self.get_child(ctx, G4VisitorParser.ExprContext)),
            bool(ctx.LBRACKET()),
        )

    def visitRuleExpr(self, ctx: G4VisitorParser.RuleExprContext):
        return RuleExpr(
            self.visitNodes(self.get_children(ctx, G4VisitorParser.RuleNameContext)),
            bool(ctx.STAR()),
            bool(ctx.QUESTION()),
        )

    def visitTokenExpr(self, ctx: G4VisitorParser.TokenExprContext):
        return TokenExpr(
            self.visitNode(self.get_child(ctx, G4VisitorParser.TokenNameContext)),
            bool(ctx.STAR()),
            bool(ctx.QUESTION()),
        )

    def visitRefExpr(self, ctx: G4VisitorParser.RefExprContext):
        return RefExpr(
            self.visitNode(self.get_child(ctx, G4VisitorParser.AttributeRefContext)),
            bool(ctx.QUESTION()),
        )

    def visitLcName(self, ctx: G4VisitorParser.LcNameContext):
        return self.visitNode(self.get_child(ctx))

    def visitUcName(self, ctx: G4VisitorParser.UcNameContext):
        return self.visitNode(self.get_child(ctx))

    def visitIdentifier(self, ctx: G4VisitorParser.IdentifierContext):
        return self.visitNode(self.get_child(ctx))

    def visitRuleName(self, ctx: G4VisitorParser.RuleNameContext):
        return self.visitNode(self.get_child(ctx))

    def visitTokenName(self, ctx: G4VisitorParser.TokenNameContext):
        return self.visitNode(self.get_child(ctx))

    def visitAttributeRef(self, ctx: G4VisitorParser.AttributeRefContext):
        return self.visitNode(self.get_child(ctx))




from collections import namedtuple

from gsl.antlr import ParseTreeVisitor
from .HedgehogTestParser import HedgehogTestParser


Message = namedtuple('Message', ('messageType', 'discriminator', 'label', 'docstring', 'fields', 'messageClasses',))
Field = namedtuple('Field', ('repeated', 'fieldType', 'name', 'label', 'languageFieldSpecs',))
Oneof = namedtuple('Oneof', ('name', 'fields',))


class HedgehogTestVisitor(ParseTreeVisitor):
    def visitExpr(self, ctx: HedgehogTestParser.ExprContext):
        return self.visitNodes(self.get_children(ctx, HedgehogTestParser.MessageContext))

    def visitMessage(self, ctx: HedgehogTestParser.MessageContext):
        return Message(
            self.visitNode(ctx.messageType),
            self.visitNode(ctx.discriminator),
            self.visitNode(ctx.label),
            self.visitNode(ctx.DOCSTRING()) if ctx.DOCSTRING() else None,
            self.visitNodes(self.get_children(ctx, HedgehogTestParser.FieldContext, HedgehogTestParser.OneofContext)),
            self.visitNodes(self.get_children(ctx, HedgehogTestParser.MessageClassContext)),
        )

    def visitField(self, ctx: HedgehogTestParser.FieldContext):
        return Field(
            bool(ctx.rep),
            self.visitNode(ctx.fieldType),
            self.visitNode(ctx.name),
            self.visitNode(ctx.label),
            self.visitNodes(self.get_children(ctx, HedgehogTestParser.LanguageFieldSpecContext)),
        )

    def visitOneof(self, ctx: HedgehogTestParser.OneofContext):
        return Oneof(
            self.visitNode(ctx.name),
            self.visitNodes(self.get_children(ctx, HedgehogTestParser.FieldContext)),
        )

    def visitQualifiedIdentifier(self, ctx: HedgehogTestParser.QualifiedIdentifierContext):
        return self.visitNodes(self.get_children(ctx, HedgehogTestParser.IdentifierContext))

    def visitIdentifier(self, ctx: HedgehogTestParser.IdentifierContext):
        return self.visitNode(self.get_child(ctx))

    def visitNumber(self, ctx: HedgehogTestParser.NumberContext):
        return self.visitNode(self.get_child(ctx))




# Generated from tests/grammar/ExprTest.g4 by ANTLR 4.7
# encoding: utf-8
from antlr4 import *
from io import StringIO
from typing.io import TextIO
import sys

def serializedATN():
    with StringIO() as buf:
        buf.write("\3\u608b\ua72a\u8133\ub9ed\u417c\u3be7\u7786\u5964\3\n")
        buf.write("%\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\3\2\3\2\3\2\3\3\3\3")
        buf.write("\3\3\7\3\21\n\3\f\3\16\3\24\13\3\3\4\3\4\3\4\7\4\31\n")
        buf.write("\4\f\4\16\4\34\13\4\3\5\3\5\3\5\3\5\3\5\5\5#\n\5\3\5\2")
        buf.write("\2\6\2\4\6\b\2\4\3\2\3\4\3\2\5\6\2#\2\n\3\2\2\2\4\r\3")
        buf.write("\2\2\2\6\25\3\2\2\2\b\"\3\2\2\2\n\13\5\4\3\2\13\f\7\2")
        buf.write("\2\3\f\3\3\2\2\2\r\22\5\6\4\2\16\17\t\2\2\2\17\21\5\6")
        buf.write("\4\2\20\16\3\2\2\2\21\24\3\2\2\2\22\20\3\2\2\2\22\23\3")
        buf.write("\2\2\2\23\5\3\2\2\2\24\22\3\2\2\2\25\32\5\b\5\2\26\27")
        buf.write("\t\3\2\2\27\31\5\b\5\2\30\26\3\2\2\2\31\34\3\2\2\2\32")
        buf.write("\30\3\2\2\2\32\33\3\2\2\2\33\7\3\2\2\2\34\32\3\2\2\2\35")
        buf.write("#\7\t\2\2\36\37\7\7\2\2\37 \5\4\3\2 !\7\b\2\2!#\3\2\2")
        buf.write("\2\"\35\3\2\2\2\"\36\3\2\2\2#\t\3\2\2\2\5\22\32\"")
        return buf.getvalue()


class ExprTestParser ( Parser ):

    grammarFileName = "ExprTest.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ "<INVALID>", "'+'", "'-'", "'*'", "'/'", "'('", "')'" ]

    symbolicNames = [ "<INVALID>", "ADD", "SUB", "MUL", "DIV", "LPAR", "RPAR", 
                      "NUMBER", "WS" ]

    RULE_expr = 0
    RULE_expr0 = 1
    RULE_summand = 2
    RULE_factor = 3

    ruleNames =  [ "expr", "expr0", "summand", "factor" ]

    EOF = Token.EOF
    ADD=1
    SUB=2
    MUL=3
    DIV=4
    LPAR=5
    RPAR=6
    NUMBER=7
    WS=8

    def __init__(self, input:TokenStream, output:TextIO = sys.stdout):
        super().__init__(input, output)
        self.checkVersion("4.7")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def expr0(self):
            return self.getTypedRuleContext(ExprTestParser.Expr0Context,0)


        def EOF(self):
            return self.getToken(ExprTestParser.EOF, 0)

        def getRuleIndex(self):
            return ExprTestParser.RULE_expr

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpr" ):
                return visitor.visitExpr(self)
            else:
                return visitor.visitChildren(self)




    def expr(self):

        localctx = ExprTestParser.ExprContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_expr)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 8
            self.expr0()
            self.state = 9
            self.match(ExprTestParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class Expr0Context(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def summand(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ExprTestParser.SummandContext)
            else:
                return self.getTypedRuleContext(ExprTestParser.SummandContext,i)


        def ADD(self, i:int=None):
            if i is None:
                return self.getTokens(ExprTestParser.ADD)
            else:
                return self.getToken(ExprTestParser.ADD, i)

        def SUB(self, i:int=None):
            if i is None:
                return self.getTokens(ExprTestParser.SUB)
            else:
                return self.getToken(ExprTestParser.SUB, i)

        def getRuleIndex(self):
            return ExprTestParser.RULE_expr0

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitExpr0" ):
                return visitor.visitExpr0(self)
            else:
                return visitor.visitChildren(self)




    def expr0(self):

        localctx = ExprTestParser.Expr0Context(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_expr0)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 11
            self.summand()
            self.state = 16
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==ExprTestParser.ADD or _la==ExprTestParser.SUB:
                self.state = 12
                _la = self._input.LA(1)
                if not(_la==ExprTestParser.ADD or _la==ExprTestParser.SUB):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 13
                self.summand()
                self.state = 18
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class SummandContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser

        def factor(self, i:int=None):
            if i is None:
                return self.getTypedRuleContexts(ExprTestParser.FactorContext)
            else:
                return self.getTypedRuleContext(ExprTestParser.FactorContext,i)


        def MUL(self, i:int=None):
            if i is None:
                return self.getTokens(ExprTestParser.MUL)
            else:
                return self.getToken(ExprTestParser.MUL, i)

        def DIV(self, i:int=None):
            if i is None:
                return self.getTokens(ExprTestParser.DIV)
            else:
                return self.getToken(ExprTestParser.DIV, i)

        def getRuleIndex(self):
            return ExprTestParser.RULE_summand

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSummand" ):
                return visitor.visitSummand(self)
            else:
                return visitor.visitChildren(self)




    def summand(self):

        localctx = ExprTestParser.SummandContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_summand)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 19
            self.factor()
            self.state = 24
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==ExprTestParser.MUL or _la==ExprTestParser.DIV:
                self.state = 20
                _la = self._input.LA(1)
                if not(_la==ExprTestParser.MUL or _la==ExprTestParser.DIV):
                    self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 21
                self.factor()
                self.state = 26
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class FactorContext(ParserRuleContext):

        def __init__(self, parser, parent:ParserRuleContext=None, invokingState:int=-1):
            super().__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return ExprTestParser.RULE_factor

     
        def copyFrom(self, ctx:ParserRuleContext):
            super().copyFrom(ctx)



    class NumberContext(FactorContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ExprTestParser.FactorContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def NUMBER(self):
            return self.getToken(ExprTestParser.NUMBER, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitNumber" ):
                return visitor.visitNumber(self)
            else:
                return visitor.visitChildren(self)


    class SubexprContext(FactorContext):

        def __init__(self, parser, ctx:ParserRuleContext): # actually a ExprTestParser.FactorContext
            super().__init__(parser)
            self.copyFrom(ctx)

        def LPAR(self):
            return self.getToken(ExprTestParser.LPAR, 0)
        def expr0(self):
            return self.getTypedRuleContext(ExprTestParser.Expr0Context,0)

        def RPAR(self):
            return self.getToken(ExprTestParser.RPAR, 0)

        def accept(self, visitor:ParseTreeVisitor):
            if hasattr( visitor, "visitSubexpr" ):
                return visitor.visitSubexpr(self)
            else:
                return visitor.visitChildren(self)



    def factor(self):

        localctx = ExprTestParser.FactorContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_factor)
        try:
            self.state = 32
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [ExprTestParser.NUMBER]:
                localctx = ExprTestParser.NumberContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 27
                self.match(ExprTestParser.NUMBER)
                pass
            elif token in [ExprTestParser.LPAR]:
                localctx = ExprTestParser.SubexprContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 28
                self.match(ExprTestParser.LPAR)
                self.state = 29
                self.expr0()
                self.state = 30
                self.match(ExprTestParser.RPAR)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx





